package mjc.gc;

import java.util.Map;
import java.util.TreeMap;

import mjc.lib.*;
import mjc.lib.classe.*;
import mjc.lib.type.*;

public class TAM extends AbstractMachine {

	private final String FAUX = "0";
	private final String VRAI = "1";
	private final String NULL = "0";

	public String getSuffixe() {
		return "tam";
	}

	private static int cpt = 0; // Pour les étiquettes

	/** Génère une étiquette. */
	private static String genEtiq() {
		return "_X_" + cpt++;
	}

	/** Génère une étiquette pour une méthode.
	 * 
	 * @param classe
	 * @param m
	 * @return classe_m.nom_m.TypeParams
	 */
	private String genEtiqMeth(String classe, METHODE m) {
		String retour = classe + "_";
		retour += m.getNom();
		for(PARAM p : m.getParams()) {
			retour += "_" + p.getType().getNom();
		}

		return retour;
	}

	private String genEtiqMeth(String classe, String ident, LPARAMS params) {
		String retour = classe + "_";
		retour += ident;
		for(PARAM p : params) {
			retour += "_" + p.getType().getNom();
		}

		return retour;
	}
	
	@Override
	public String genererTDV(TDC classes) {
		String retour = "; Table des virtuelles : tableau d'indirection des méthodes des classes\n";
		// On va parcourir la liste des classes par ordre de deplacement croissant
		// deplacement = @ dans la pile
		
		// On crée donc la liste des classes rangée par ordre de déplacement
		
		// Remarque: On ne peut pas utiliser de liste, car il y a des "trous"
		Map<Integer, CLASSE> tdc = new TreeMap<Integer, CLASSE>();
		for(CLASSE c : classes.values()) {
			int dep = c.getDeplacement();
			tdc.put(dep, c);
		}
		
		for(CLASSE c : tdc.values()) {
				retour += "\t; CLASSE " + c.getNom() + "\n";
				for(METHODE m : c.getMethodes()) {
					// On ne doit pas prendre en compte les constructeurs ni main
					String nom = m.getNom();
					if(!(nom.equals(c.getNom()) || nom.equals("main"))) {
						retour += "\tLOADA " + genEtiqMeth(m.getClasse(), m) + "\n";
					}
				}
			}
		return retour + "\n\n\n";
	}

	@Override
	public String genererAppelMain(TDC classes) {
		String retour = "";
		for(CLASSE c : classes.values()) {
			for(METHODE m : c.getMethodes()) {
				if (m.getNom().equals("main") && (m.getParams().size() == 0)
									&& (m.getType() instanceof ENTIER)) {
					retour = "\tCALL (LB) " + genEtiqMeth(c.getNom(), m) + "\n"; 
				}
			}
		}
		if (!retour.equals("")) {
			retour = "\t; Appel de la fonction main\n" + retour;
			retour += "\t LOADL \"FIN DU PROGRAMME !\\n\"\n";
			retour += "\tSUBR SOut\n";
			retour += "\tLOADL \"VALEUR DE RETOUR : \"\n";
			retour += "\tSUBR SOut\n";
			retour += "\tSUBR IOut\n";
			// Flush du buffer
			retour += "\tLOADL \"\\n\"\n";
			retour += "\tSUBR SOut\n";
			retour += "\tHALT";
		} else {
			retour = null;
		}
		return retour;  
	}
	
	@Override
	//TODO: Free sur les pointeurs
	public String genFree(int taille) {
		String retour = "";
		if (taille != 0) {
			retour +=  "\t;Liberation de " + taille + " mot" + (taille > 1 ? "s" : "");
			retour += "\n\tPOP (0) " + taille +"\n";
		}
		return retour;
	}
	
	@Override
	public String genMethode(String classe, METHODE methode, String code) {
		String etiq = genEtiqMeth(classe, methode);
		String retour =  "; Méthode "+methode.getNom() + " de la classe " + classe + "\n";
		retour += etiq + "\n";
		retour += code;
		if(methode.getType().equals(new VIDE())) {
			retour += "\tRETURN (0) " + (methode.getParams().getTaille() + 1) + "\n\n"; 
		}
		return retour;
	}
	
	@Override
	public String genRetour(METHODE m, TDS locales) {
		LPARAMS params = m.getParams();
		DTYPE type = m.getType();
		int taille_param = params.getTaille() + 1; // Ne pas oublier this !
		if(m.getNom().equals("main")) {
			taille_param--; //Pas de this dans un main
		}
		int taille_locales = locales.getTaille();

		String retour = "";
		
		if (taille_locales > 1) {
			// Si = 1 : C'est la valeur à retourner
			retour += "\t; Nettoyage des locales : " + taille_locales + "\n";			
			retour += "\t; Il faut conserver la valeur à retourner !\n";
			retour += "\tPOP (" + type.getTaille() + ") " + taille_locales + "\n";
		}
		retour += "\t; Retour de la fonction : " + taille_param + " parametres\n";
		retour += "\tRETURN (" + type.getTaille() + ") " + taille_param  + "\n\n";

		return retour;
	}
	
	@Override
	public String genConstructeur(String classe, METHODE constructeur, String code) {
		String etiq = genEtiqMeth(classe, constructeur);
		String retour = "; Constructeur de la classe " + classe + "\n";
		retour += etiq + "\n";
		// On appelle le constructeur par défaut
		retour += "\t; Appel au constructeur par défaut\n";
		retour += "\tCALL (LB) " + classe + "_defaut\n";
		retour += code;
		if (!code.equals("")) { // (Toute) petite optimisation :D
			retour += "\tLOAD (1) 3[LB] ; On rempile l'@ du pointeur à retourner\n";
		}
		retour += "\tRETURN (1) " + constructeur.getParams().size() + "\n\n";
		return retour;
	}
	
	@Override
	public String genConstructeurDefaut(CLASSE c) {
		String retour = "\t; Constructeur par défaut de la classe " + c.getNom() + "\n";
		retour += c.getNom() + "_defaut\n";
		
		int taille = c.getAttributs().getTaille();
		taille++; // Pour la TDV en début d'objet.
		
		int indice_TDV = c.getDeplacement();

		retour += "\t; Allocation de l'espace mémoire\n";
		retour += "\tLOADL " + taille + " ;Taille de l'objet + TDV\n";
		retour += "\tSUBR MAlloc ; Allocation \n";
		retour += "\tLOADL " + indice_TDV + " ; @TDV de la classe\n";
		retour += "\tLOAD (1) -2[ST] ; On récupère l'@ fournie par le malloc\n";
		retour += "\tSTOREI (1) ; Et on affecte la valeur de la TDV dans l'objet crée\n";
		
		retour += "\t; Initialisation des attributs\n";
		
		for(PARAM p : c.getAttributs()) {
			DTYPE t = p.getType();
			// On initialise chacun des attributs de type entier à 42
			if(t instanceof ENTIER) {
				retour += "\t\t; Initialisation de " + p.getNom() + " (type entier) \n";
				retour += "\tLOADL 42 ; Valeur par défaut: 42\n";
			} else if (t instanceof BOOLEEN) {
				retour += "\t\t; Initialisation de " + p.getNom() + " (type booléen) \n";
				retour += "\tLOADL " + FAUX + " ; Valeur par défaut: FAUX\n";
			} else { // Les pointeurs sont initialisés à NULL
				retour += "\t\t; Initialisation de " + p.getNom() + " (pointeur) \n";
				retour += "\tLOADL " + NULL +" ; Valeur par défaut: NULL\n";
			}
			// XXX : Non optimisé :(
			retour += "\t; On récupère l'@ de this\n";
			retour += "\tLOAD (1) 3[LB]\n";
			retour += "\tLOADL " + p.getDeplacement() + "\n";
			retour += "\tSUBR IAdd\n"; // ou l'inverse ?
			retour += "\tSTOREI (1) ; Et on sauve le tout !\n";
		}
		retour += "\tRETURN (1) 0\n\n";
		return retour;
	}

	@Override
	public String genererCodeNPE() {
		String retour = "\t\n; En cas d'exception \n";
		retour += "_NPE_\n";
		retour += "\tLOADL \"Null Pointer Exception\"\n";
		retour += "\tSUBR SOut\n";
		retour += "\tHALT\n\n";
		return retour;
	}

	@Override
	public String genAppelMeth(CLASSE classe, METHODE m) {
		
		String retour = "\t; Appel de la méthode " + classe.getNom() + "::" + m.getNom() + "\n";

		int deplacement = classe.getMethodes().indice(m);
		int taille_param = m.getParams().getTaille();
		
		retour += "\tLOADL 0 ; Pour le CALLI (cf doc de tam)\n";
		// Le pointeur sur lequel on applique la méthode est empilé
		retour += "\tLOAD (1) " + -(taille_param + 2) + "[ST] ; On récupère la valeur du pointeur\n";
		retour += "\tLOADI (1) ; On récupère l'@ de la TDV\n";
		if (deplacement > 0) {
			retour += "\tLOADL " + deplacement + " ; Id de la méthode dans la TDV\n";
			retour += "\tSUBR IAdd\n";
		}
		retour += "\t; On obtient l'adresse réelle de la méthode\n";
		retour += "\tLOADI (1) ; Que l'on va chercher et mettre en tête de pile\n";
//		retour += "\tLOADL 30\nPOP (0) 1\n"; //Pour debug
		retour += "\tCALLI ; Et on appelle la méthode.. OUF\n";
		
		return retour;
	}

	@Override
	public String genEcritureMem(String adresse, String expression, DTYPE type) {
		String retour = "\t; Ecriture mémoire\n\t; -> Calul de l'expression\n";
		retour += expression;
		retour += "\t; -> Calcul de l'adresse\n";
		retour += adresse;
		retour += "\tSTOREI (" + type.getTaille() + ")\n";
		return retour;
	}
	
	@Override
	public String genLectureMem(String code, DTYPE type) {
		String retour = "\t; Lecture mémoire\n";
		retour += code;
		retour += "\tLOADI (" + type.getTaille() + ")\n";
		return retour;
	}
	
	@Override
	public String genLectureMem(String code, DTYPE type, boolean estValeur) {
		String retour = code;
		if(!estValeur) {
			retour =  genLectureMem(code, type);
		}
		return retour;
	}
	
	@Override
	public String genAffect(VARIABLE variable) {
		return "\tSTORE (" + variable.getType().getTaille() + ") "+
		variable.getDep() + "[LB]\n";
	}

	@Override
	public String genDecl(VARIABLE var) {
		String retour = "";
		DTYPE t = var.getType();
		if(t instanceof POINTEUR) {
			retour += "\t;Declaration d'un pointeur\n";
			retour += "\tLOADL " + NULL + "\n";
		} else if(t instanceof ENTIER) {
			retour += "\t;Declaration d'un entier\n";
			retour += "\tLOADL 42\n";
		} else if(t instanceof BOOLEEN){
			retour += "\t;Declaration d'un booléen\n";
			retour += "\tLOADL " + FAUX + "\n";
		}

		return retour;
	}
	
	@Override
	public String genAppelConstructeur(DTYPE type, METHODE m, String code) {
		String retour = "\t;Appel au constructeur :\n";
		CLASSE classe = ((POINTEUR)type).getClasse();
		
		// On charge les paramètres du constructeur
		retour += code;
		
		// Et on génère l'appel :)
		String etiq = "";
		if(m == null) {
			// C'est l'appel au constucteur par défaut
			etiq = classe.getNom() + "_defaut";
		} else {
			etiq = genEtiqMeth(classe.getNom(), classe.getNom(), m.getParams());
		}
		retour += "\tCALL (LB) " + etiq + " ; Appel du constructeur\n";

		return retour;
	}
	
	@Override
	public String genVerifNPE() {
		String retour = "\t; Vérification d'une NPE\n";
		retour += "\tLOAD (1) -1[ST]\n";
		retour += "\tJUMPIF (" + NULL + ") _NPE_\n";
		return retour;
	}
	
	@Override
	public String genAdresseAttribut(PARAM attribut) {
		// Suppose que le pointeur est empilé en tête
		String retour = "\t; @Attribut : " + attribut.getNom() + "\n";
		retour += "\tLOADL " + attribut.getDeplacement() + "\n";
		retour += "\tSUBR IAdd\n"; // Ou l'inverse ?
		return retour;
	}

	@Override
	public String genAdresseParam(PARAM param) {
		String retour = "\t; @Parametre : " + param.getNom() + "\n";
		retour += "\tLOADA " + - (param.getDeplacement() + 1) + " [LB]\n";
		return retour;
	}

	@Override
	public String genAdresseVariable(VARIABLE var) {
		String retour = "\t; @Variable\n";
		retour += "\tLOADA " + var.getDep() + "[LB]\n";
		return retour;
	}
	
	@Override
	public String genAccesVariable(VARIABLE var) {
		String retour = "\t; Valeur variable\n";
		retour += "\tLOAD (" + var.getType().getTaille() +") "
				+ var.getDep() + "[LB]\n";
		return retour;
		
	}

	@Override
	public String genAccesParam(PARAM param) {
		String retour = "\t; Valeur param : " + param.getNom() + "\n";
		retour += "\tLOAD (" + param.getType().getTaille() + ") "
					+ -(param.getDeplacement() + 1) + "[LB]\n";
		return retour;
	}

	@Override
	public String genAccesAttribut(PARAM attribut) {
		// Suppose que this est empilé en tête
		String retour = "\t; Valeur attribut : " + attribut.getNom() + "\n";
		retour += "\tLOADL " + attribut.getDeplacement() + "\n";
		retour += "\tSUBR IAdd \n";
		retour += "\tLOADI (" + attribut.getType().getTaille() +")\n";
		return retour;
	}
	
	@Override
	public String genEmpilerThis(METHODE m) {
		// Cas nominal : nombre de paramètre + 1 : le this
		int dep = -(m.getParams().getTaille() + 1);
		if(m.getClasse().equals(m.getNom())) {
			// C'est un constructeur
			//On doit aller chercher @this après LB
			dep = 3;
		}
		String retour = "\t; On empile this\n";
		retour += "\tLOAD (1) " + dep + "[LB]\n";
		return retour;
	}
	
	@Override
	public String genAffichage(String code, DTYPE type) {
		String retour = "\t; Affichage\n";
		retour += code;
		// On affiche les booléens en tant que tel
		if (type instanceof BOOLEEN) {
			retour += "\tSUBR BOut\n";
		} else { // Tout le reste sera en tant qu'entier (@ des pointeurs par ex)
			retour += "\tSUBR IOut\n";
		}
		retour += "\t; On vide le buffer\n";
		retour += "\tLOADL \"\\n\"\n";
		retour += "\tSUBR SOut\n";
		return retour;
	}	
	
	@Override
	public String genSi(String condition, String alors, String sinon) {
		String etiq_sinon = genEtiq(); String etiq_fin = genEtiq();

		String retour = condition;
		retour += "\t; Test condition (SI)\n";
		retour += "\tJUMPIF (" + FAUX + ") ";

		if(sinon.equals("")) { 		// Dans le cas d'un si/alors
			retour += etiq_fin + "\n";
			retour += "\t; BLOC ALORS\n";
			retour += alors;
		} else {
			retour += etiq_sinon + "\n";
			retour += "\t; BLOC ALORS\n";
			retour += alors;
			retour += "\tJUMP " + etiq_fin + "\n";
			retour += "\t; BLOC SINON\n";
			retour += etiq_sinon +"\n" + sinon;
		}

		retour += "\t; FIN SI\n";
		retour += etiq_fin + "\n";

		return retour;
	}	
	
	@Override
	public String genTq(String condition, String boucle) {
		String etiq_bcle = genEtiq(); String etiq_fin = genEtiq();
		String retour = etiq_bcle;
		retour += "\t; Test condition (TANT QUE)\n";
		retour += "\tJUMPIF (" + FAUX + ") " + etiq_fin + "\n";
		retour += boucle + "\n";
		retour += "\t; Fin TQ\n";
		retour += "\tJUMP " + etiq_bcle + "\n";
		retour += etiq_fin + "\n";
		return retour;
	}
	
	@Override
	public String genNull() {
		return "\tLOADL " + NULL + "\n";
	}
	
	@Override
	public String genEntier(String valeur) {
		return "\tLOADL " + valeur + "\n";
	}
	
	@Override
	public String genFaux() {
		return "\tLOADL " + FAUX + "\n";
	}
	
	@Override
	public String genVrai() {
		return "\tLOADL " + VRAI + "\n";
	}
	
	@Override
	public String genAddition() {
		return "\tSUBR IADD\n";
	}

	@Override
	public String genDiv() {
		return "\tSUBR IDiv\n";
	}

	@Override
	public String genEg() {
		return "\tSUBR IEq\n";
	}

	@Override
	public String genEt() {
		return "\tSUBR BAnd\n";
	}

	@Override
	public String genInf() {
		return "\tSUBR ILss\n";
	}

	@Override
	public String genInfEg() {
		return "\tSUBR ILeq\n";
	}
	
	@Override
	public String genMod() {
		return "\tSUBR IMod\n";
	}

	@Override
	public String genMoins() {
		return "\tSUBR ISub\n";
	}

	@Override
	public String genMult() {
		return "\tSUBR IMul\n";
	}

	@Override
	public String genNeq() {
		return "\tSUBR INeqn";
	}

	@Override
	public String genNon() {
		return "\tSUBR BNeg\n";
	}

	@Override
	public String genOpp() {
		return "\tSUBR INeg\n";
	}

	@Override
	public String genOu() {
		return "\tSUBR BOr\n";
	}	

	@Override
	public String genSup() {
		return "\tSUBR IGtr\n";
	}

	@Override
	public String genSupEg() {
		return "\tSUBR IGeq\n";
	}
}
